/* code.c       near optimal hierarchical encoding of type hierarchies        */
/* author:      Andreas Krall                                                 */
/* last change: October 15 1996                                               */
/* All rights reserved, see license                                           */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PTRINTTYPE long

#ifndef SETTYPESIZE

#ifdef __alpha
#define SETTYPESIZE 64
#else
#define SETTYPESIZE 32
#endif

#endif /* SETTYPESIZE */

#if SETTYPESIZE==64
#define SETWORDS   8
#else
#define SETWORDS  16
#endif


#define SETTYPE    long
#define SHIFT      ((sizeof(SETTYPE)>>3)+5)  /* sizeof(SETTYPE) in {4, 8, 16} */
#define MASK       ((sizeof(SETTYPE)<<3)-1)

#define MAXSETSIZE (SETWORDS<<SHIFT)


#define HASHMAX     512                 /* must be power of two               */
#define TEXTMAX   32000                 /* must be <= ALLOCSIZE               */
#define ALLOCSIZE 32000                 /* size for memory allocation         */

#define TopNode       0
#define EndOfFile    -1
#define SuperKeyWord -2
#define NameKeyWord  -3

typedef SETTYPE SET[SETWORDS];
typedef SETTYPE *SETPTR;

typedef struct MEMORY *MEMPTR;

typedef struct MEMORY {
	char   *firstfree;
	char   *endptr;
	MEMPTR next;
	char   mem[ALLOCSIZE];
	} MEMORY;

typedef struct NODE *NODEPTR;
typedef struct NOLI *NOLIPTR;
typedef struct CNLI *CNLIPTR;

typedef struct NODE {              /* node table element                      */
	int     id;                    /* node id                                 */
	int     needs_bit;             /* true if mi node nees his own bit        */
	int     bit_pos;               /* bit position of the node                */
	SET     bit_set;               /* code set of hierarchical encoding       */
	int     par_count;             /* number of parents                       */
	int     child_count;           /* number of children                      */
	int     cost;                  /* cost of node, sum of children needs_bit */
	int     anc_count;             /* number of ancestors                     */
	int     conflict_count;        /* number of conflicting nodes (degree)    */
	int     desc_count;            /* number of mi descendants                */
	int     help_count;            /* used during balancing and colouring     */
	int     level;                 /* level in type hierarchy                 */
	int     path_length;           /* path length from root to leaf via node  */
	int     path_leaf;             /* leaf node of path length                */
	int     cur_dist_root;         /* current distance to root node           */
	int     cur_dist_leaf;         /* current distance to leaf node           */
	int     min_dist_root;         /* minimal distance to root node           */
	int     min_dist_leaf;         /* minimal distance to leaf node           */
	int     lower_path;            /* lower path length via node              */
	int     lower_bound;           /* lower bound of different bits needed    */
	int     lower_leaf;            /* leaf node of lower bound                */
	NOLIPTR parent_list;           /* list of parents                         */
	CNLIPTR child_list;            /* list of childs                          */
	NOLIPTR conflict_list;         /* list of conflicting nodes               */
	SETPTR  ancestors;             /* bitvector of all ancestors              */
	SETPTR  descendants;           /* bitvector of mi descendants             */
	SETPTR  conflicts;             /* bitvector of all conflicts              */
	char    *name;                 /* pointer to the characters of node       */
	NODEPTR hash_next;             /* next node with same hash value          */
	NODEPTR next;                  /* next node                               */
	NOLIPTR prev;                  /* prev node list in uncoloured list       */
	int     col_count[MAXSETSIZE]; /* occurence count of colors in conflicts  */
	                               /* is set by one less for fast comparisons */
	} NODE;

typedef struct NOLI {
	NODEPTR  node;
	NOLIPTR  next;
	} NOLI;

typedef struct CNLI {
	int      bucket;
	NODEPTR  node;
	CNLIPTR  next;
	} CNLI;

typedef struct SORTNL *SORTPTR;

typedef struct SORTNL {
	NOLIPTR first;
	NOLIPTR last;
	} SORTNL;

typedef struct SORTCL *CLSOPTR;

typedef struct SORTCL {
	int     empty_desc;
	CNLIPTR first;
	CNLIPTR last;
	SETPTR  descendants;
	} SORTCL;

typedef int (*INTNOLIPTRFUNC)(NODEPTR node);

enum CharClasses {      /* partitioning of characters     */
	UA,                 /* upper case alphabetic          */
	LA,                 /* lower case alphabetic          */
	UL,                 /* underline character            */
	DI,                 /* digit, UA, UL and LA < DI !!   */
	SQ,                 /* "'" single quote               */
	LP,                 /* "(" left paranthesis           */
	RP,                 /* ")" right paranthesis          */
	LB,                 /* "[" left bracket               */
	RB,                 /* "]" right bracket              */
	LC,                 /* "{" left curly bracket         */
	RC,                 /* "}" right curly bracket        */
	VB,                 /* "|" vertical bar               */
	CM,                 /* "," comma                      */
	PT,                 /* "." point                      */
	SC,                 /* ";" semi colon                 */
	CL,                 /* ":" colon                      */
	LT,                 /* "<" less than                  */
	EQ,                 /* "=" equal                      */
	GT,                 /* ">" greater than               */
	AD,                 /* "+" add                        */
	SB,                 /* "-" sub                        */
	ML,                 /* "*" mul                        */
	DV,                 /* "/" div                        */
	CR,                 /* carriage return                */
	LF,                 /* line feed                      */
	BS,                 /* blank space                    */
	IC,                 /* invalid character              */
	EF                  /* end of file                    */
	};

static int cclass[256] = {
/* nul soh stx etx eot enq ack bel bs  ht  nl  vt  np  cr  so  si  */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,

/* dle dc1 dc2 dc3 dc4 nak syn etb can em  sub esc fs  gs  rs  us  */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,

/* sp  !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /   */
   BS, BS, BS, BS, UA, UA, UA, BS, BS, BS, UA, UA, BS, UA, UA, UA,

/* 0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?   */
   UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, BS, BS, UA, UA, UA, BS,

/* @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O   */
   BS, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA,

/* P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _   */
   UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, BS, BS, BS, BS, UA,

/* `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o   */
   BS, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA,

/* p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~   del */
   UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, UA, BS, BS, BS, BS, BS,

   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS
};


static PTRINTTYPE alignmentmask[] = {
	0, -1, -2, -4, -4, -8, -8, -8, -8, -16, -16, -16, -16, -16, -16, -16, -16
	};

static FILE    *infile;             /* source program file                    */
static int     cread;               /* last character read                    */
static int     cc;                  /* character class of cread               */
static NODEPTR hashtab[HASHMAX];    /* hash table for node search             */
static NODEPTR root = NULL;         /* root node                              */
static NODEPTR last = NULL;         /* last node in node list                 */
static int     nodecount = 0;       /* count of nodes in node table           */
static int     orgnodecount = 0;    /* count of nodes without balancing       */
static int     free_nolicount = 0;  /* count of noli's in free list           */
static int     mi_count = 0;        /* count of nodes with multiple inh       */
static int     nodecountmax = 0;    /* count of nodes in node table           */
static int     conflict_nodes = 0;  /* count of nodes in conflict graph       */
static int     node_set_size;       /* size of node set                       */
static char    *text;               /* character space for names              */
static char    *endtext;            /* first free position in text            */
static char    *maxtext;            /* maximum position in text               */
static int     edges = 0;           /* number of edges in relation            */
static int     max_par_count = 0;   /* maximum number of parents              */
static int     max_child_count = 0; /* maximum number of childs               */
static int     max_level = 0;       /* maximum level in type hierarchy        */
static int     bit_count = 0;       /* bit size of transitive closure         */
static int     matrix_bytes = 0;    /* number of bytes for binary matrix      */
static int     anc_list_bytes = 0;  /* number of bytes for ancestor list      */
static int     min_bytes = 0;       /* number of bytes for minimal repr.      */
static double  avg_degree = 0;      /* average degree of conflict graph       */
static int     mem_use = 0;         /* size of allocoated memory              */
static int     max_bit_pos = -1;    /* biggest bit position                   */
static int     maxdegree = 0;       /* biggest degree                         */
static int     dist_max;            /* maximum of root-leaf distance          */
static int     split_bound = 5;     /* bound for binary splitting             */
static NODEPTR biggest_lower_bound_leaf_node;
static NODEPTR biggest_path_length_leaf_node;
static int     all_mi_get_bit = 0;  /* true, if all mi nodes get an own bit   */
static int     not_early = 1;       /* true, if not early presplitting        */
static int     not_bucket = 1;      /* true, if not bucket presplitting       */
static int     not_bucket_v = 1;    /* true, if not bucket presplitting       */
static int     z_split = 0;         /* new splitting with old propagation     */
static SETPTR  anc_set;             /* set of ancester nodes, used by lattice */
static SETPTR  node_set;            /* set of nodes, used by balancing        */
static SETPTR  split_set;           /* set of nodes, used by split_tree       */
static SETPTR  si_nodes;            /* set of single inheritance nodes        */
static NODEPTR *mi_nodes = NULL;    /* list of multiple inheritance nodes     */
static NODEPTR *mi_list = NULL;     /* tmp list of multiple inheritance nodes */
static NOLIPTR uncoloured = NULL;   /* list of uncoloured nodes               */
static SORTPTR linear_sort = NULL;  /* array for linear sorting of node lists */
static CLSOPTR child_sort = NULL;   /* array for sorting of child lists       */



#define AdjustPtr(p, a) (char*)((((PTRINTTYPE)(p))+(a)-1)&alignmentmask[a])

#define NextChar()      if((cread=getc(infile))==EOF)cc=EF;else cc=cclass[cread]

#define SetBit(sptr,nr) (sptr)[(nr)>>SHIFT]|=((SETTYPE)1)<<((nr)&MASK)
#define TstBit(sptr,nr) (((sptr)[(nr)>>SHIFT]&(((SETTYPE)1)<<((nr)&MASK)))!=0)
#define NotTstBit(sptr,nr) (((sptr)[(nr)>>SHIFT]&(((SETTYPE)1)<<((nr)&MASK)))==0)
#define TstBitShiftMask(sptr,s,m) (((sptr)[s]&(m))!=0)
#define NotTstBitShiftMask(sptr,s,m) (((sptr)[s]&(m))==0)

static void clear_set(SET s) {
int i;

for (i = 0; i < SETWORDS; i++)
	s[i] = 0;
}


static int is_empty_set(SET s) {
int i;

for (i = 0; i < SETWORDS; i++)
	if (s[i] != 0)
		return 0;
return 1;
}


static void union_set(SET source, SET dest) {
int i;

for (i = 0; i < SETWORDS; i++)
	dest[i] |= source[i];
}


static void section_set(SET source1, SET source2, SET dest) {
int i;

for (i = 0; i < SETWORDS; i++)
	dest[i] = source1[i] & source2[i];
}


static int equal_set(SET source1, SET source2) {
int i;

for (i = 0; i < SETWORDS; i++)
	if (source1[i] != source2[i])
		return 0;
return 1;
}


static void copy_set(SET source, SET dest) {
int i;

for (i = 0; i < SETWORDS; i++)
	dest[i] = source[i];
}


static int a_and_b_is_equal_b_set(SET source1, SET source2) {
int i;

for (i = 0; i < SETWORDS; i++)
	if ((source1[i] & source2[i]) != source2[i])
		return 0;
return 1;
}


static int a_and_b_is_equal_b_set_1(SET source1, SET source2) {
int i;

#if SETWORDS >  1
if ((source1[ 1] & source2[ 1]) != source2[ 1]) return 0;
#endif
#if SETWORDS >  2
if ((source1[ 2] & source2[ 2]) != source2[ 2]) return 0;
#endif
#if SETWORDS >  3
if ((source1[ 3] & source2[ 3]) != source2[ 3]) return 0;
#endif
#if SETWORDS >  4
if ((source1[ 4] & source2[ 4]) != source2[ 4]) return 0;
#endif
#if SETWORDS >  5
if ((source1[ 5] & source2[ 5]) != source2[ 5]) return 0;
#endif
#if SETWORDS >  6
if ((source1[ 6] & source2[ 6]) != source2[ 6]) return 0;
#endif
#if SETWORDS >  7
if ((source1[ 7] & source2[ 7]) != source2[ 7]) return 0;
#endif
#if SETWORDS > 8
for (i = 8; i < SETWORDS; i++)
	if ((source1[i] & source2[i]) != source2[i])
		return 0;
#endif

return 1;
}


static void clear_node_set(SETPTR s) {
int i;

for (i = 0; i < node_set_size; i++)
	s[i] = 0;
}


static void copy_node_set(SETPTR source, SETPTR dest) {
int i;

for (i = 0; i < node_set_size; i++)
	dest[i] = source[i];
}


static void union_node_set(SETPTR source, SETPTR dest) {
int i;

for (i = 0; i < node_set_size; i++)
	dest[i] |= source[i];
}


static void section_node_set(SETPTR source1, SETPTR source2, SETPTR dest) {
int i;

for (i = 0; i < node_set_size; i++)
	dest[i] = source1[i] & source2[i];
}


static int equal_node_set(SETPTR source1, SETPTR source2) {
int i;

for (i = 0; i < node_set_size; i++)
	if (source1[i] != source2[i])
		return 0;
return 1;
}


static int overlapped_node_set(SETPTR source1, SETPTR source2) {
int i;

for (i = 0; i < node_set_size; i++)
	if (source1[i] & source2[i])
		return 1;
return 0;
}


static void memory_error(void) {
printf("Out of Memory!!!\n");
exit(1);
}


static void file_error(char *name) {
printf("Error opening file %s!!!\n", name);
exit(2);
}


static void syntax_error(char *err_text) {
printf("Syntax error, %s, found %s!!\n", err_text, text);
exit(3);
}


static void system_error(char *err_text) {
printf("System error, %s!!\n", err_text);
exit(3);
}


static void *new_global(int size, int alignment) {
char *retptr;
static MEMPTR memptr = NULL;

#ifdef DEBUG
if ((size + 15) > ALLOCSIZE)
	memory_error();
#endif
if ((memptr == NULL) ||
    (((retptr = AdjustPtr(memptr->firstfree, alignment)) + size) > memptr->endptr))
    {
	if ((memptr = malloc(sizeof(MEMORY))) == NULL)
		memory_error();
	mem_use += sizeof(MEMORY);
	memptr->endptr = memptr->mem+ALLOCSIZE;
	retptr = AdjustPtr(memptr->mem, alignment);
	}
memptr->firstfree = retptr + size;
return retptr;
}


static NOLIPTR new_node_list(NODEPTR node, NOLIPTR next) {
static NOLIPTR free_list = NULL;
static int     free_count = 0;

free_count--;
free_list++;
if (free_count <= 0) {
	free_count = 500;
	if ((free_list = malloc(sizeof(NOLI) * free_count)) == NULL)
		memory_error();
	mem_use += sizeof(NOLI) * free_count;
	}
free_list->node = node;
free_list->next = next;
return free_list;
}


static CNLIPTR new_child_node_list(NODEPTR node, CNLIPTR next) {
static CNLIPTR free_list = NULL;
static int     free_count = 0;

free_count--;
free_list++;
if (free_count <= 0) {
	free_count = 100;
	if ((free_list = malloc(sizeof(CNLI) * free_count)) == NULL)
		memory_error();
	mem_use += sizeof(CNLI) * free_count;
	}
free_list->bucket = -1;
free_list->node = node;
free_list->next = next;
return free_list;
}


static SETPTR  new_node_set(void) {
static SETPTR  free_list = NULL;
static int     free_count = 0;

free_count--;
if (free_count <= 0) {
	if (node_set_size == 0)
		system_error("can't allocate node set before knowing number of nodes");
	if (free_list == NULL) 
		free_count = nodecount + 20;
	else
		free_count = 20;
	free_list = malloc(sizeof(SETTYPE) * node_set_size * free_count);
	if (free_list == NULL)
		memory_error();
	mem_use += sizeof(SETTYPE) * node_set_size * free_count;
	free_list -= node_set_size;
	}
free_list += node_set_size;
return free_list;
}


static NODEPTR new_node(void) {
NODEPTR node;

if ((node = malloc(sizeof(NODE))) == NULL)
	memory_error();
mem_use += sizeof(NODE);
node->id = nodecount++;
node->needs_bit = 1;
node->bit_pos = -1;
node->par_count = 0;
node->child_count = 0;
node->anc_count = 1;
node->conflict_count = 0;
node->desc_count = 0;
node->help_count = 0;
node->lower_bound = 0;
node->path_length = 0;
node->parent_list = NULL;
node->child_list = NULL;
node->conflict_list = NULL;
node->name = "";
node->hash_next = NULL;
node->next = NULL;
if (root == NULL)
	root = node;
else
	last->next = node;
last = node;
if ((nodecountmax != 0) && (nodecount >= nodecountmax))
	system_error("not enough node reserve");
return node;
}


static NODEPTR find_or_insert_node(void) {
int     i;
NODEPTR node;

i = (text[0] + (endtext[-1] << 2) + 
    (text[(endtext - text) >> 1] << 4)) & (HASHMAX - 1);
if (++endtext >= maxtext)
	memory_error();
node = hashtab[i];
while (node != NULL)
	if (strcmp(node->name, text) == 0)
		return node;
	else
		node = node->hash_next;
node = new_node();
node->name = strcpy(new_global(endtext - text, 1), text);
node->hash_next = hashtab[i];
hashtab[i] = node;
return node;
}


static int next_name(void) {
while (cc == BS) {
	NextChar();
	}
if (cc == EF)
	return EndOfFile;
endtext = text;
do {
	*endtext++ = cread;
	if (endtext >= maxtext)
		memory_error();
	NextChar();
	} while (cc == UA);
*endtext = 0;
if (strcmp("super", text) == 0)
	return SuperKeyWord;
if (strcmp("name", text) == 0)
	return NameKeyWord;
return 0;
}


static void init(char *filename) {
int i;

if ((text = malloc(sizeof(char) * TEXTMAX)) == NULL)
	memory_error();
mem_use += sizeof(char) * TEXTMAX;
maxtext = text + TEXTMAX;

i = HASHMAX;
while (--i >= 0)
	hashtab[i] = NULL;

endtext = text;
*endtext++ = '@';
*endtext++ = '$';
*endtext++ = 'T';
*endtext++ = 'o';
*endtext++ = 'p';
*endtext++ = '!';
*endtext = 0;
if ((root = find_or_insert_node()) == NULL)
	system_error("top node initialisation");
if ((infile = fopen(filename, "r")) == NULL)
	file_error(filename);
NextChar();
}


static void enter_edge(NODEPTR child, NODEPTR parent) {
NOLIPTR  p;

parent->child_list = new_child_node_list(child, parent->child_list);;
parent->child_count++;
child->parent_list = new_node_list(parent, child->parent_list);
child->par_count++;
edges++;
if (child->par_count > max_par_count)
	max_par_count = child->par_count;
if (parent->child_count > max_child_count)
	max_child_count = parent->child_count;
}


static void read_data(void) {
int     i;
NODEPTR parent, child;

i = next_name();
while (i != EndOfFile) {
	if (i != SuperKeyWord)
		syntax_error("super expected");
	i = next_name();
	if (i == EndOfFile)
		syntax_error("Unexpected end of file");
	parent = find_or_insert_node();
	i = next_name();
	if (i != NameKeyWord)
		syntax_error("name expected");
	i = next_name();
	if (i == EndOfFile)
		syntax_error("Unexpected end of file");
	child = find_or_insert_node();
	i = next_name();
	if (parent == child)
		enter_edge(child, root);
	else
		enter_edge(child, parent);
	}
}


static void build_line(SETPTR ancestors, NODEPTR node, NOLIPTR parents) {
int id;

while (parents != NULL) {
	id = parents->node->id;
	if (NotTstBit(ancestors, id)) {
		bit_count++;
		SetBit(ancestors, id);
		node->anc_count++;
		build_line(ancestors, node, parents->node->parent_list);
		}
	parents = parents->next;
	}
}


static void build_descendants(int id, NOLIPTR parents) {
NODEPTR node;

while (parents != NULL) {
	node = parents->node;
	if (NotTstBit(node->descendants, id)) {
		SetBit(node->descendants, id);
		node->desc_count++;
		build_descendants(id, node->parent_list);
		}
	parents = parents->next;
	}
}


static void check_for_minimal_rep(NODEPTR node) {
NOLIPTR xl, yl, pl;
CNLIPTR cl;
NODEPTR x, y;

if (node->par_count <= 1)
	return;
xl = node->parent_list;
while (xl != NULL) {
	x  = xl->node;
	xl = xl->next;
	for (yl = node->parent_list; yl != NULL;) {
		y  = yl->node;
		yl = yl->next;
		if (y == x)
			continue;
		if (NotTstBit(y->ancestors, x->id))
			continue;
/*		printf("deleting edge between child %3d %17s and parent %3d %17s because of %3d %17s!!\n",
		       node->id, node->name, x->id, x->name, y->id, y->name); */
		edges--;
		node->par_count--;
		pl = node->parent_list;
		if (pl->node == x)
			node->parent_list = pl->next;
		else {
			while (pl->next->node != x)
				pl = pl->next;
			pl->next = pl->next->next;
			}
		x->child_count--;
		cl = x->child_list;
		if (cl->node == node)
			x->child_list = cl->next;
		else {
			while (cl->next->node != node)
				cl = cl->next;
			cl->next = cl->next->next;
			}
		break;
		}
	}
}


static void mi_parents(NOLIPTR parents) {
while (parents != NULL) {
	if (parents->node->bit_pos == -1) {
		mi_parents(parents->node->parent_list);
		parents->node->bit_pos = -2;
		if (parents->node->par_count > 1)
			*mi_list++ = parents->node;
		}
	parents = parents->next;
	}
}


static void mi_childs(CNLIPTR childs) {
while (childs != NULL) {
	if (NotTstBit(node_set, childs->node->id)) {
		SetBit(node_set, childs->node->id);
		if (childs->node->bit_pos == -1) {
			mi_parents(childs->node->parent_list);
			childs->node->bit_pos = -2;
			if (childs->node->par_count > 1)
				*mi_list++ = childs->node;
			}
		mi_childs(childs->node->child_list);
		}
	childs = childs->next;
	}
}


static void build_mi_nodes(void) {
NODEPTR node;

mi_list = mi_nodes;
for (node = root; node != NULL; node = node->next)
	node->bit_pos = -1;
clear_node_set(node_set);
mi_childs(root->child_list);
mi_list = mi_nodes;
}


static int has_mi_conflict(SETPTR ancx, SETPTR ancy) {
int     i;
SETTYPE xsi, *si;

si = si_nodes;
i = node_set_size;
while (--i > 0) {
	si++;
	ancy++;
	xsi = *ancx++ & *si;
	if ((*ancy & xsi) != xsi)
		return 0;
	}
return 1;
}


static void check_mi_conflict(void) {
NODEPTR node, x, y, *mix, *miy;
SETTYPE xsi0, idmask, *ancx;
int     idshift, mi_conflict_count;

mi_conflict_count = 0;
for (mix = mi_nodes; (x = *mix) != NULL; mix++)
	x->needs_bit = 0;
for (mix = mi_nodes; (x = *mix) != NULL; mix++) {
	if (x->needs_bit)
		continue;
	idshift = x->id >> SHIFT;
	idmask = ((SETTYPE) 1) << (x->id & MASK);
	ancx = x->ancestors;
	xsi0 = ancx[0] & si_nodes[0];
	ancx++;
	for (miy = mi_nodes; (y = *miy) != NULL; miy++) {
		if (TstBitShiftMask(y->ancestors, idshift, idmask))
			continue;
		if ((y->ancestors[0] & xsi0) != xsi0)
			continue;
		if (!has_mi_conflict(ancx, y->ancestors))
			continue;
		x->needs_bit = 1;
		SetBit(si_nodes, x->id);
		mi_conflict_count++;
		break;
		}
	}
printf("%d mi nodes out of %d need a bit\n", mi_conflict_count, mi_count);
}


static void compute_level_childs(int level, CNLIPTR childs) {
NODEPTR node;

while (childs != NULL) {
	node = childs->node;
	if (node->level < level + 1) {
		node->level = level + 1;
		compute_level_childs(level + 1, node->child_list);
		}
	childs = childs->next;
	}
}


static int compute_max_level(void) {
NODEPTR node;

for (node = root->next; node != NULL; node = node->next)
	node->level = 0;
root->level = 1;
compute_level_childs(1, root->child_list);
max_level = 1;
for (node = root->next; node != NULL; node = node->next)
	if (max_level < node->level)
		max_level = node->level;
return max_level;
}


static void build_matrix(void) {
NODEPTR node;
NOLIPTR nl;
CNLIPTR cl;
SETPTR  ancestors, descendants, conflicts;
int     i, leaf_node_count, old_edges, all_parents, max_ancestors, all_ancestors;

for (node = root->next; node != NULL; node = node->next)
	if (node->par_count == 0) {
		printf(" Undeclared top node %3d %s !!!\n", node->id, node->name);
		enter_edge(node, root);
		}
if (root->child_count == 1) {
	if (root->next == root->child_list->node)
		root = root->next;
	else {
		node = root->child_list->node;
		while (node->next != NULL)
			node = node->next;
		node->next = root->next;
		root = root->child_list->node;
		while (node->next != root)
			node = node->next;
		last = node;
		last->next = NULL;
		}
	root->par_count = 0;
	root->parent_list = NULL;
	
	for (node = root; node != NULL; node = node->next)
		node->id--;
	nodecount--;
	edges--;
	}
orgnodecount = nodecount;
for (i = 0, node = root; node != NULL; node = node->next)
	if (node->child_count > 2)
		i += node->child_count - 2;
node_set_size = (nodecount + i + MASK) >> SHIFT;
nodecountmax = (node_set_size << SHIFT) - 1;
printf("%d nodes reserved for balancing\n", i);
if ((linear_sort = malloc(sizeof(SORTNL) * (nodecount + i + 1))) == NULL)
	memory_error();
mem_use += sizeof(SORTNL) * (nodecount + i);
leaf_node_count = 0;
anc_set    = new_node_set();
node_set   = new_node_set();
split_set  = new_node_set();
si_nodes   = new_node_set();
clear_node_set(si_nodes);
for (node = root; node != NULL; node = node->next) {
	ancestors = node->ancestors = new_node_set();
	descendants = node->descendants = new_node_set();
	conflicts = node->conflicts = new_node_set();
	for (i = 0; i < node_set_size; i++) {
		*ancestors++ = 0;
		*descendants++ = 0;
		*conflicts++ = 0;
		}
	build_line(node->ancestors, node, node->parent_list);
	SetBit(node->ancestors, node->id);
	SetBit(node->conflicts, node->id);
	}
old_edges = edges;
mi_count = 0;
for (node = root; node != NULL; node = node->next) {
	check_for_minimal_rep(node);
	if (node->par_count <= 1) {
		SetBit(si_nodes, node->id);
		}
	else
		mi_count++;
	}
if (old_edges != edges)
	printf("%d edges deleted !!\n", old_edges - edges);
if ((mi_nodes = malloc(sizeof(NODEPTR) * mi_count + 1)) == NULL)
	memory_error();
mem_use += sizeof(NODEPTR) * mi_count + 1;
build_mi_nodes();
if (!all_mi_get_bit)
	check_mi_conflict();
root->needs_bit = 0;
anc_list_bytes = 0;
min_bytes = 0;
max_ancestors = 0;
all_ancestors = 0;
all_parents = 0;
for (node = root; node != NULL; node = node->next) {
	all_parents += node->par_count;
	if (max_ancestors < node->anc_count)
		max_ancestors = node->anc_count;
	all_ancestors += node->anc_count;
	for (nl = node->parent_list, i = 0; nl != NULL; nl = nl->next)
		if (i < nl->node->anc_count)
			i = nl->node->anc_count;
	anc_list_bytes += node->anc_count - i;
	min_bytes += (node->par_count + 1) << 2;
	if (node->child_count == 0) {
		leaf_node_count++;
		node->cost = 0;
		}
	else {
		for (cl = node->child_list, i = 0; cl != NULL; cl = cl->next)
			i += cl->node->needs_bit;
		node->cost = i;
		}
	if (node->par_count > 1)
		build_descendants(node->id, node->parent_list);
	}
if ((child_sort = malloc(sizeof(SORTCL) * (max_child_count + 1))) == NULL)
	memory_error();
mem_use += sizeof(SORTCL) * (nodecount + i);
for (i = 0; i <= max_child_count; i++)
	child_sort[i].descendants = new_node_set();
anc_list_bytes = anc_list_bytes << 3;
printf("%d nodes out of %d are leaf nodes\n", leaf_node_count, nodecount);
printf("%d nodes and %d edges read in\n", nodecount, edges);
printf("%d nodes are nodes with more than one parent\n", mi_count);
printf("The depth of the hierarchy is %d\n", compute_max_level());
printf("The maximum number of parents per node is %d\n", max_par_count);
printf("The average number of parents per node is %6.4f\n",
        (double) all_parents / nodecount);
printf("The maximum number of ancestors per node is %d\n", max_ancestors);
printf("The average number of ancestors per node is %7.4f\n",
        (double) all_ancestors / nodecount);
printf("The maximum number of childs per node is %d\n", max_child_count);
if (mi_count < (nodecount >> 5))
	split_bound = 4;
}


static void test_line(NOLIPTR parents) {

while (parents != NULL) {
	if (NotTstBit(anc_set, parents->node->id)) {
		SetBit(anc_set, parents->node->id);
		test_line(parents->node->parent_list);
		}
	parents = parents->next;
	}
}


static int test_parents(NODEPTR node, NOLIPTR parents) {
CNLIPTR childs;
int     parcnt;

parcnt = 0;
while (parents != NULL) {
	parcnt++;
	childs = parents->node->child_list;
	while (childs != NULL) {
		if (childs->node == node)
			break;
		childs = childs->next;
		}
	if (childs == NULL)
		printf("Child node %d missing in parent node %d\n",
			node->id, parents->node->id);
	parents = parents->next;
	}
return parcnt;
}


static void test_childs(NODEPTR node, CNLIPTR childs) {
NOLIPTR parents;

while (childs != NULL) {
	parents = childs->node->parent_list;
	while (parents != NULL) {
		if (parents->node == node)
			break;
		parents = parents->next;
		}
	if (parents == NULL)
		printf("Parent node %d missing in child node %d\n",
			node->id, childs->node->id);
	childs = childs->next;
	}
}


static void test_matrix(void) {
NODEPTR node;
CNLIPTR childs;
int     childcnt;

for (node = root; node != NULL; node = node->next) {
	if (test_parents(node, node->parent_list) != node->par_count)
		printf("Wrong parent count in node %d %s !!\n", node->id, node->name);
	test_childs(node, node->child_list);
	clear_node_set(anc_set);
	SetBit(anc_set, node->id);
	test_line(node->parent_list);
	if (!equal_node_set(anc_set, node->ancestors))
		printf("Wrong ancestors in node %d %s !!\n", node->id, node->name);
	childcnt = 0;
	for (childs = node->child_list; childs != NULL; childs = childs->next)
		childcnt++;
	if (childcnt != node->child_count)
		printf("Wrong child count in node %d %s !!\n", node->id, node->name);
	}
printf("\nmatrix tested !!!!!!!!!!!!!!!!!!\n");
}


static void compute_min_dist_root(int length, CNLIPTR childs) {
NODEPTR node;
int     node_length;

while (childs != NULL) {
	node = childs->node;
	node_length = length + node->needs_bit;
	if (node->min_dist_root < node_length) {
		node->min_dist_root = node_length;
		compute_min_dist_root(node_length, node->child_list);
		}
	childs = childs->next;
	}
}


static void compute_min_dist_leaf(int dist, NOLIPTR parents) {
NOLIPTR pars;

while (parents != NULL)
	if (parents->node->min_dist_leaf < dist) {
		parents->node->min_dist_leaf = dist;
		for (pars = parents->next; pars != NULL; pars = pars->next) {
			if (pars->node->min_dist_leaf < dist) {
				pars->node->min_dist_leaf = dist;
				compute_min_dist_leaf(dist + pars->node->needs_bit,
		                      pars->node->parent_list);
				}
			}
		dist += parents->node->needs_bit;
		parents = parents->node->parent_list;
		}
	else
		parents = parents->next;
}


static void compute_bound_down(int length, int cost, CNLIPTR childs) {
NODEPTR node;
int     node_length;

do {
	node = childs->node;
	node_length = length + node->needs_bit;
	if (cost > node->needs_bit)
		node_length++;
	if (node->lower_bound < node_length) {
		node->lower_bound = node_length;
		if (node->child_list != NULL)
			compute_bound_down(node_length, node->cost, node->child_list);
		}
	childs = childs->next;
	} while (childs != NULL);
}


static void lower_path_parents(int bound, int needs_bit, NOLIPTR parents) {
int     node_bound;
NOLIPTR childs;
NODEPTR node;

while (parents != NULL) {
	node = parents->node;
	node_bound = node->needs_bit;
	if (node->cost > needs_bit)
		node_bound++;
	node_bound += bound;
	if (node_bound > node->help_count) {
		if (node_bound > node->lower_path)
			node->lower_path = node_bound;
		node->help_count = node_bound;
		lower_path_parents(node_bound, node->needs_bit, node->parent_list);
		}
	parents = parents->next;
	}
}


static void store_lower_bound(int bound, int leaf, NOLIPTR parents) {
NOLIPTR pars;
NODEPTR node;

while (parents != NULL) {
	node = parents->node;
	if (node->lower_bound < bound) {
		node->lower_bound = bound;
		node->lower_leaf = leaf;
		pars = parents->next;
		parents = node->parent_list;
		while (pars != NULL) {
			node = pars->node;
			if (node->lower_bound < bound) {
				node->lower_bound = bound;
				node->lower_leaf = leaf;
				store_lower_bound(bound, leaf, node->parent_list);
				}
			pars = pars->next;
			}
		}
	else
		parents = parents->next;
	}
}


static void store_path_length(int bound, int leaf, NOLIPTR parents) {
NOLIPTR pars;
NODEPTR node;

while (parents != NULL) {
	node = parents->node;
	if (node->path_length < bound) {
		node->path_length = bound;
		node->path_leaf = leaf;
		pars = parents->next;
		parents = node->parent_list;
		while (pars != NULL) {
			node = pars->node;
			if (node->path_length < bound) {
				node->path_length = bound;
				node->path_leaf = leaf;
				store_path_length(bound, leaf, node->parent_list);
				}
			pars = pars->next;
			}
		}
	else
		parents = parents->next;
	}
}


static void compute_bounds(void) {
int     max_lower_bound;
NODEPTR node;

max_lower_bound = 0;
for (node = root; node != NULL; node = node->next) {
/*
	node->min_dist_root = 0;
	node->min_dist_leaf = -1;
	node->lower_path = 0;
*/
	node->lower_bound = -2;
	}
/*
compute_min_dist_root(0, root->child_list);
dist_max = 0;
*/
compute_bound_down(-1, root->cost, root->child_list);
for (node = root; node != NULL; node = node->next)
	if (node->child_count == 0) {
/*
		if (dist_max < node->min_dist_root)
			dist_max = node->min_dist_root;
		node->min_dist_leaf = 0;
		compute_min_dist_leaf(node->needs_bit, node->parent_list);
*/
		node->lower_bound++;
		store_lower_bound(node->lower_bound, node->id, node->parent_list);
		if (node->lower_bound > max_lower_bound) {
			max_lower_bound = node->lower_bound;
			biggest_lower_bound_leaf_node = node;
			}
		}
}


static void compute_dist_root(int dist, CNLIPTR childs) {
NODEPTR node;

do {
	node = childs->node;
	if (node->cur_dist_root < dist) {
		node->cur_dist_root = dist;
		if (node->child_list != NULL)
			compute_dist_root(dist + node->cost, node->child_list);
		}
	childs = childs->next;
	} while (childs != NULL);
}


static void compute_dist_leaf(int dist, int leaf, NOLIPTR parents) {
NOLIPTR pars;

while (parents != NULL) {
	if (parents->node->cur_dist_leaf < (dist + parents->node->cost)) {
		parents->node->cur_dist_leaf = dist + parents->node->cost;
		parents->node->path_leaf = leaf;
		for (pars = parents->next; pars != NULL; pars = pars->next) {
			if (pars->node->cur_dist_leaf < (dist + pars->node->cost)) {
				pars->node->cur_dist_leaf = dist + pars->node->cost;
				pars->node->path_leaf = leaf;
				compute_dist_leaf(dist + pars->node->cost, leaf,
				                  pars->node->parent_list);
				}
			}
		dist += parents->node->cost;
		parents = parents->node->parent_list;
		}
	else
		parents = parents->next;
	}
}


static void compute_path_length(void) {
int     max_path_length;
NODEPTR node;

max_path_length = 0;
for (node = root; node != NULL; node = node->next) {
	node->cur_dist_root = 0;
	node->path_length = 0;
	}
root->cur_dist_root = 0;
if (root->child_list != NULL)
	compute_dist_root(root->cost, root->child_list);
for (node = root; node != NULL; node = node->next)
	if (node->child_count == 0) {
		if (max_path_length < node->cur_dist_root) {
			max_path_length = node->cur_dist_root;
			biggest_path_length_leaf_node = node;
			}
		node->path_length = node->cur_dist_root;
		node->path_leaf = node->id;
		store_path_length(node->path_length, node->id, node->parent_list);
		}
}


static void new_compute_path_length(void) {
int     max_path_length;
NODEPTR node;

max_path_length = 0;
for (node = root; node != NULL; node = node->next) {
	node->cur_dist_root = 0;
	node->cur_dist_leaf = -1;
	}
root->cur_dist_root = 0;
if (root->child_list != NULL)
	compute_dist_root(root->cost, root->child_list);
for (node = root; node != NULL; node = node->next)
	if (node->child_count == 0) {
		if (max_path_length < node->cur_dist_root) {
			max_path_length = node->cur_dist_root;
			biggest_path_length_leaf_node = node;
			}
		node->cur_dist_leaf = 0;
		node->path_leaf = node->id;
		compute_dist_leaf(0, node->id, node->parent_list);
		}
/* printf("leaf node %3d has maximal path length %d(%d)\n",
        biggest_path_length_leaf_node->id, max_path_length, root->cur_dist_leaf);
*/
for (node = root; node != NULL; node = node->next)
	node->path_length = node->cur_dist_root + node->cur_dist_leaf;
}


static void sort_upper(void) {
NODEPTR node;
int     min_bound;
CNLIPTR childs, sorted, list, next, min_list, min_prev;

for (node = root; node != NULL; node = node->next)
	if (node->child_count > 1) {
		childs = node->child_list;
		sorted = NULL;
		while (childs != NULL) {
			min_list = list = childs;
			min_bound = list->node->path_length;
			while ((next = list->next) != NULL) {
				if (next->node->path_length < min_bound) {
					min_bound = next->node->path_length;
					min_list = next;
					min_prev = list;
					}
				list = next;
				}
			if (min_list == childs)
				childs = min_list->next;
			else
				min_prev->next = min_list->next;
			min_list->next = sorted;
			sorted = min_list;
			}
		node->child_list = sorted;
/*
		printf("Node %d childs are ", node->id);
		print_sorted(sorted);
		printf("\n");
*/
		}
}


static void update_ancestors(CNLIPTR nodelist, int id) {
while (nodelist != NULL) {
	SetBit(nodelist->node->ancestors, id);
	update_ancestors(nodelist->node->child_list, id);
	nodelist = nodelist->next;
	}
}


static void replace_parent(CNLIPTR cl, NODEPTR oldparent, NODEPTR newparent) {
NOLIPTR parents;

while (cl != NULL) {
	parents = cl->node->parent_list;
	while (parents->node != oldparent) {
		parents = parents->next;
		}
	parents->node = newparent;
	cl = cl->next;
	}
}


static void update_cost(NODEPTR node) {
int     i;
CNLIPTR cl;

for (cl = node->child_list, i = 0; cl != NULL; cl = cl->next)
	i += cl->node->needs_bit;
node->cost = i;
}


static int split_tree_rec(NOLIPTR nodelist, SETPTR  ancestors) {
CNLIPTR childs, next, anc_list, big_list, small_list;
NODEPTR node, newnode;
int     bound, anc, rest, big_rest, small_rest, i;
int testsmall_rest;

while (nodelist != NULL) {
/*
	printf("Node %d has upper bound %d and %d childs\n", nodelist->node->id,
	        nodelist->node->path_length, nodelist->node->.child_count);
*/
	node = nodelist->node;
	bound = node->path_length;
	if (TstBit(split_set, node->id)) {
		nodelist = nodelist->next;
		continue;
		}
	if (nodelist->node->child_count > 2) {
		childs = node->child_list;
		anc = rest = big_rest = 0;
		while (childs != NULL) {
			if (TstBit(ancestors, childs->node->id))
				anc++;
			else {
				rest++;
				if (childs->node->path_length > bound)
					break;
				if (childs->node->path_length == bound)
					big_rest++;
				}
			childs = childs->next;
			}
		if (childs != NULL) {
			nodelist = nodelist->next;
			continue;
			}
		small_rest = rest - big_rest;

/*		printf("childcount = %d, rest = %d, big_rest = %d, anc = %d\n",
		        nodelist->node->child_count, rest, big_rest, anc); */

		if ((rest > 1) && ((big_rest == 0) ||
		    ((anc > 3) && (rest > 3)) || (rest > split_bound))) {
/* the following was best for self (96 bits) */
/*		if ((rest > 1) && ((big_rest == 0) ||
		    ((anc > 3) && (rest > 3)) || (rest > 5))) { */
/*		if ((rest > 1) && ((big_rest == 0) ||
		    ((anc > 2) && (rest > 2)) || (rest > 3))) { */
/*		if ((rest > 1) && ((big_rest == 0) || (anc > 1) || (rest > 2))) { */
/*		if ((rest > 1) && ((big_rest == 0) || (small_rest > 1))) { */
			anc_list = small_list = big_list = NULL;
			testsmall_rest = 0;
			childs = node->child_list;
			while (childs != NULL) {
				next = childs->next;
				if (TstBit(ancestors, childs->node->id)) {
					childs->next = anc_list;
					anc_list = childs;
					}
				else
					if (childs->node->path_length == bound) {
						childs->next = big_list;
						big_list = childs;
						}
					else {
						childs->next = small_list;
						small_list = childs;
						testsmall_rest++;
						}
				childs = next;
				}
			newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			if (big_rest == 0) {
				newnode->child_count = small_rest;
				newnode->child_list = small_list;
				update_ancestors(small_list, newnode->id);
				replace_parent(small_list, node, newnode);
				node->child_count = anc + 1;
				node->child_list = new_child_node_list(newnode, anc_list);
/*				printf("Node %4d has been split in lists of length %d(%d) and %d(%d), one node added\n",
				        node->id, (int) (anc + 1),
				        (int) anc_list->node->path_length, (int) small_rest,
				        (int) small_list->node->path_length); */
				update_cost(node);
				update_cost(newnode);
				return 1;
				}
			if (small_rest > 1) {
				newnode->child_count = small_rest;
				newnode->child_list = small_list;
				update_ancestors(small_list, newnode->id);
				replace_parent(small_list, node, newnode);
				node->child_count = anc + big_rest + 1;
				node->child_list = new_child_node_list(newnode, anc_list);
				i = anc;
				while (--i > 0)
					anc_list = anc_list->next;
				anc_list->next = big_list;
/*				printf("Node %4d has been split in lists of length %d + %d + 1 and %d, one node added\n",
				        node->id, anc, big_rest, small_rest); */
				update_cost(node);
				update_cost(newnode);
				return 1;
				}
			/* anc > 1, big_rest > 0, small_rest <= 1 */
			if (small_list == NULL)
				small_list = big_list;
			else
				small_list->next = big_list;
			if ((i = (rest - anc) >> 1) > 0) {
				anc += i;
				rest -= i;
				next = anc_list;
				anc_list = small_list;
				while (--i > 0)
					small_list = small_list->next;
				childs = small_list;
				small_list = small_list->next;
				childs->next = next; /* old anc_list */
				}
			update_ancestors(small_list, newnode->id);
			replace_parent(small_list, node, newnode);
			newnode->child_count = rest;
			newnode->child_list = small_list;
			node->child_count = 2;
			node->child_list = new_child_node_list(newnode,
			                                 new_child_node_list(NULL, NULL));
			node->child_list->next->node = newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			newnode->child_count = anc;
			newnode->child_list = anc_list;
			update_ancestors(anc_list, newnode->id);
			replace_parent(anc_list, node, newnode);
/*			printf("Node %4d has been split in lists of length %d and %d, two nodes added\n",
			        node->id, (int) anc, (int) rest); */
			update_cost(node);
			update_cost(node->child_list->node);
			update_cost(node->child_list->next->node);
			return 1;
			}
		}
	if (split_tree_rec(nodelist->node->parent_list, ancestors))
		return 1;
	nodelist = nodelist->next;
	}
return 0;
}


static int split_tree(NOLIPTR nodelist, SETPTR  ancestors) {
clear_node_set(split_set);
return split_tree_rec(nodelist, ancestors);
}

static int new_split_tree_rec(NOLIPTR nodelist, SETPTR  ancestors) {
CNLIPTR childs, next, anc_list, big_list, small_list;
NODEPTR node, newnode;
int     bound, leaf, anc, rest, big_rest, small_rest, i;

while (nodelist != NULL) {
/*
	printf("Node %d has upper bound %d and %d childs\n", nodelist->node->id,
	        nodelist->node->path_length, nodelist->node->.child_count);
*/
	node = nodelist->node;
	bound = node->path_length;
	leaf = node->path_leaf;
	if (TstBit(split_set, node->id)) {
		nodelist = nodelist->next;
		continue;
		}
	SetBit(split_set, node->id);
	if (nodelist->node->child_count > 2) {
		anc = rest = big_rest = 0;
		for (childs = node->child_list; childs != NULL; childs = childs->next) {
			if (childs->node->path_length > bound)
				break;
			if (childs->node->path_length < bound)
				rest++;
			else
				if (childs->node->path_leaf == leaf)
					anc++;
				else
					big_rest++;
			}
		if (childs != NULL) {
			nodelist = nodelist->next;
			continue;
			}
		small_rest = rest;
		rest += big_rest;

/*		printf("childcount = %d, rest = %d, big_rest = %d, anc = %d\n",
		        nodelist->node->child_count, rest, big_rest, anc); */
/*		if ((anc > 3) && (rest <= 2))
			printf("Lost splitting chance in node %4d(%d, %d) !!!!!!!!!!!!!!!!!!\n",
			        node->id, anc, rest); */
		if ((rest > 1) && ((small_rest > 1) ||
		    ((anc > 2) && (rest > 2)) || (rest > split_bound))) {
/* the following was best for self (96 bits) */
/*		if ((rest > 1) && ((big_rest == 0) ||
		    ((anc > 3) && (rest > 3)) || (rest > 5))) { */
/*		if ((rest > 1) && ((big_rest == 0) ||
		    ((anc > 2) && (rest > 2)) || (rest > 3))) { */
/*		if ((rest > 1) && ((big_rest == 0) || (anc > 1) || (rest > 2))) { */
/*		if ((rest > 1) && ((big_rest == 0) || (small_rest > 1))) { */
			anc_list = small_list = big_list = NULL;
			childs = node->child_list;
			while (childs != NULL) {
				next = childs->next;
				if (childs->node->path_length < bound) {
						childs->next = small_list;
						small_list = childs;
						}
				else
					if (childs->node->path_leaf == leaf) {
						childs->next = anc_list;
						anc_list = childs;
						}
					else {
						childs->next = big_list;
						big_list = childs;
						}
				childs = next;
				}
			newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			if ((big_rest == 0) && (small_rest > 1)) {
				newnode->child_count = small_rest;
				newnode->child_list = small_list;
				update_ancestors(small_list, newnode->id);
				replace_parent(small_list, node, newnode);
				node->child_count = anc + 1;
				node->child_list = new_child_node_list(newnode, anc_list);
/*				printf("Node %4d has been 1. split in lists of length %d(%d) and %d(%d), one node added\n",
				        node->id, (int) (anc + 1),
				        (int) anc_list->node->path_length, (int) small_rest,
				        (int) small_list->node->path_length); */
				update_cost(node);
				update_cost(newnode);
				return 1;
				}
			if ((big_rest > 0) && (small_rest > 1)) {
				newnode->child_count = small_rest;
				newnode->child_list = small_list;
				update_ancestors(small_list, newnode->id);
				replace_parent(small_list, node, newnode);
				node->child_count = anc + big_rest + 1;
				node->child_list = new_child_node_list(newnode, anc_list);
				i = anc;
				while (--i > 0)
					anc_list = anc_list->next;
				anc_list->next = big_list;
/*				printf("Node %4d has been split in lists of length %d + %d + 1 and %d, one node added\n",
				        node->id, anc, big_rest, small_rest); */
				update_cost(node);
				update_cost(newnode);
				return 1;
				}
			/* anc > 0, big_rest > 0, small_rest <= 1 */
			if (small_list == NULL)
				small_list = big_list;
			else
				small_list->next = big_list;
			if ((i = (rest - anc) >> 1) > 0) {
				anc += i;
				rest -= i;
				next = anc_list;
				anc_list = small_list;
				while (--i > 0)
					small_list = small_list->next;
				childs = small_list;
				small_list = small_list->next;
				childs->next = next; /* old anc_list */
				}
			update_ancestors(small_list, newnode->id);
			replace_parent(small_list, node, newnode);
			newnode->child_count = rest;
			newnode->child_list = small_list;
			node->child_count = 2;
			node->child_list = new_child_node_list(newnode,
			                                 new_child_node_list(NULL, NULL));
			node->child_list->next->node = newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			newnode->child_count = anc;
			newnode->child_list = anc_list;
			update_ancestors(anc_list, newnode->id);
			replace_parent(anc_list, node, newnode);
/*			printf("Node %4d has been split in lists of length %d and %d, two nodes added\n",
			        node->id, (int) anc, (int) rest); */
			update_cost(node);
			update_cost(node->child_list->node);
			update_cost(node->child_list->next->node);
			return 1;
			}
		}
	if (new_split_tree_rec(nodelist->node->parent_list, ancestors))
		return 1;
	nodelist = nodelist->next;
	}
return 0;
}


static int new_split_tree(NOLIPTR nodelist, SETPTR  ancestors) {
clear_node_set(split_set);
return new_split_tree_rec(nodelist, ancestors);
}

static void fast_splits(void) {
int     i, bound;
NODEPTR node;

compute_path_length();
for (;;) {
	if (!split_tree(biggest_path_length_leaf_node->parent_list,
	    biggest_path_length_leaf_node->ancestors))
		break;
	compute_path_length();
/*	printf("The lower bound is %d, the upper bound is %d\n",
	        root->lower_bound, root->path_length); */
	}
bound = root->lower_bound;
for (node = root; node != NULL; node = node->next) {
	if (node->child_count != 0)
		continue;
	while (node->path_length > bound)
	    if (split_tree(node->parent_list, node->ancestors))
			compute_path_length();
		else
			break;
	}
}


static void new_splits(void) {
int     i, bound;
NODEPTR node;

if (z_split)
	compute_path_length();
else
	new_compute_path_length();
for (;;) {
	if (!new_split_tree(biggest_path_length_leaf_node->parent_list,
	    biggest_path_length_leaf_node->ancestors))
		break;
	if (z_split)
		compute_path_length();
	else
		new_compute_path_length();
/*	printf("The lower bound is %d, the upper bound is %d\n",
	        root->lower_bound, root->path_length); */
	}
bound = root->lower_bound;
for (node = root; node != NULL; node = node->next) {
	if (node->child_count != 0)
		continue;
	while (node->path_length > bound)
	    if (new_split_tree(node->parent_list, node->ancestors))
			if (z_split)
				compute_path_length();
			else
				new_compute_path_length();
		else
			break;
	}
}


static void sort_childs_pre_split(void) {
NODEPTR node, help;
int     i, j, maxbucket, maxbound, maxleaf;
CNLIPTR childs, list, maxlist;
SETPTR  desc;

for (node = root; node != NULL; node = node->next) if (node->child_count > 2) {
	for (childs = node->child_list; childs->next != NULL; childs = childs->next) {
		maxlist = childs;
		maxbound = childs->node->lower_bound;
		maxleaf = childs->node->lower_leaf;
		for (list = childs->next; list != NULL; list = list->next)
			if ((maxbound < list->node->lower_bound) ||
			    ((maxbound == list->node->lower_bound) &&
			     (maxleaf < list->node->lower_leaf))) {
				maxbound = list->node->lower_bound;
				maxleaf = list->node->lower_leaf;
				maxlist = list;
				}
		help = childs->node;
		childs->node = maxlist->node;
		maxlist->node = help;
		}
	maxbucket = 0;
	for (childs = node->child_list; childs != NULL; childs = list) {
		list = childs->next;
		if (childs->node->desc_count == 0) {
			child_sort[maxbucket].empty_desc = 1;
			child_sort[maxbucket].first = child_sort[maxbucket].last = childs;
			maxbucket++;
			continue;
			}
		desc = childs->node->descendants;
		for (i = 0; i < maxbucket; i++)
			if (!child_sort[i].empty_desc &&
			     overlapped_node_set(child_sort[i].descendants, desc)) {
				union_node_set(desc, child_sort[i].descendants);
				child_sort[i].last = child_sort[i].last->next = childs;
				for (j = i + 1; j < maxbucket; j++)
					if (!child_sort[j].empty_desc &&
					     overlapped_node_set(child_sort[j].descendants, desc)) {
						union_node_set(child_sort[j].descendants,
						               child_sort[i].descendants);
						child_sort[i].last->next = child_sort[j].first;
						child_sort[i].last = child_sort[j].last;
						child_sort[j].first = NULL;
						child_sort[j].empty_desc = 1;
						}
				break;
				}
		if (i == maxbucket) {
			child_sort[i].empty_desc = 0;
			child_sort[i].first = child_sort[i].last = childs;
			copy_node_set(desc, child_sort[i].descendants);
			maxbucket++;
			}
		}
	childs = NULL;
	i = maxbucket;
	maxbucket = 0;
	while (--i >= 0) 
		if (child_sort[i].first != NULL) {
			child_sort[i].last->next = NULL;
			for (list = child_sort[i].first; list != NULL; list = list->next) {
				list->bucket = maxbucket;
/*				printf("child %4d in bucket %3d has path length %2d \n", 
				       list->node->id, list->bucket, list->node->lower_bound);*/
				}
			maxbucket++;
			child_sort[i].last->next = childs;
			childs = child_sort[i].first;
			}
	node->child_list = childs;
	for (childs = node->child_list; childs->next != NULL; childs = childs->next) {
		maxlist = childs;
		maxbucket = childs->bucket;
		maxbound = childs->node->lower_bound;
		maxleaf = childs->node->lower_leaf;
		for (list = childs->next; list != NULL; list = list->next)
			if ((maxbucket < list->bucket) ||
			    ((maxbucket == list->bucket) &&
			     (maxbound < list->node->lower_bound)) ||
			    ((maxbucket == list->bucket) &&
			     (maxbound == list->node->lower_bound) &&
			     (maxleaf < list->node->lower_leaf))) {
				maxbucket = list->bucket;
				maxbound = list->node->lower_bound;
				maxleaf = list->node->lower_leaf;
				maxlist = list;
				}
		maxlist->bucket = childs->bucket;
		childs->bucket = maxbucket;
		help = childs->node;
		childs->node = maxlist->node;
		maxlist->node = help;
		}
	}
}


static void pre_split(void) {
NODEPTR node, newnode;
int     powertwo, power_rest, power_diff, i, counta, countb, lbound, bucket;
CNLIPTR lista, listb;

sort_childs_pre_split();
for (node = root; node != NULL; node = node->next) {
	if (node->child_count > 13) {
		powertwo = 8;
		while ((powertwo << 1) < node->child_count)
			powertwo = powertwo << 1;
		power_rest = node->child_count - powertwo;
		power_diff = powertwo - power_rest;
		listb = node->child_list;
		lbound = listb->node->lower_bound;
		for (i = 1; i < power_rest; i++) {
			bucket = listb->bucket;
			listb = listb->next;
			if (listb->bucket != bucket)
				lbound = listb->node->lower_bound;
			}
		lista = listb;
		counta = power_rest;
		for (i = 0; i <= power_diff; i++) {
			if ((listb->bucket != listb->next->bucket) &&
			    (lbound != listb->next->node->lower_bound))
				break;
			listb = listb->next;
			}
		listb = lista;
		if (not_bucket && (i <= power_diff))
			for (i = 0; i <= power_diff; i++) {
				if ((listb->bucket != listb->next->bucket) &&
				    (lbound != listb->next->node->lower_bound)) {
					lista = listb;
					counta += i;
					break;
					}
				listb = listb->next;
				}
		else {
			for (i = 0; i <= power_diff; i++) {
				if (listb->bucket != listb->next->bucket)
					break;
				listb = listb->next;
				}
			listb = lista;
			if (not_bucket_v && (i <= power_diff))
				for (i = 0; i <= power_diff; i++) {
					if (listb->bucket != listb->next->bucket) {
						lista = listb;
						counta += i;
						break;
						}
					listb = listb->next;
					}
			else {
				for (i = 0; i <= power_diff; i++) {
					if (listb->node->lower_bound >
					    listb->next->node->lower_bound)
						break;
					listb = listb->next;
					}
				listb = lista;
				if (not_early && (i <= power_diff))
					for (i = 0; i <= power_diff; i++) {
						if (listb->node->lower_bound >
						    listb->next->node->lower_bound) {
							lista = listb;
							counta += i;
							break;
							}
						listb = listb->next;
						}
				else
					for (i = 0; i <= power_diff; i++) {
						if (listb->node->lower_leaf !=
						    listb->next->node->lower_leaf) {
							lista = listb;
							counta += i;
							break;
							}
						listb = listb->next;
						}
				}
			}
		listb = lista->next;
		lista->next = NULL;
		lista = node->child_list;
		countb = node->child_count - counta;
			newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			update_ancestors(lista, newnode->id);
			replace_parent(lista, node, newnode);
			newnode->child_count = counta;
			newnode->child_list = lista;
			node->child_count = 2;
			node->child_list = new_child_node_list(newnode,
			                                 new_child_node_list(NULL, NULL));
			node->child_list->next->node = newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts = new_node_set();
			newnode->ancestors = new_node_set();
			clear_node_set(newnode->conflicts);
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			newnode->child_count = countb;
			newnode->child_list = listb;
			update_ancestors(listb, newnode->id);
			replace_parent(listb, node, newnode);
			update_cost(node);
			update_cost(node->child_list->node);
			update_cost(node->child_list->next->node);
/*			printf("Node %d has been presplit in lists of length %d and %d, two nodes added\n",
			        node, counta, countb); */
		}
	}
}


static void eight_split(int splitval) {
NODEPTR node, newnode;
int     powertwo, power_rest, power_diff, i, counta, countb;
CNLIPTR lista, listb;

if (splitval < 5)
	splitval = 5;
for (node = root; node != NULL; node = node->next) {
	if (node->child_count >= splitval) {
		powertwo = 4;
		power_rest = node->child_count - powertwo;
		power_diff = powertwo - power_rest;
		listb = node->child_list;
		for (i = 1; i < power_rest; i++)
			listb = listb->next;
		lista = listb;
		counta = power_rest;
		listb = lista->next;
		lista->next = NULL;
		lista = node->child_list;
		countb = node->child_count - counta;
			newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts =
				new_node_set();
			clear_node_set(newnode->conflicts);
			newnode->ancestors =
				new_node_set();
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			update_ancestors(lista, newnode->id);
			replace_parent(lista, node, newnode);
			newnode->child_count = counta;
			newnode->child_list = lista;
			node->child_count = 2;
			node->child_list = new_child_node_list(newnode,
			                                 new_child_node_list(NULL, NULL));
			node->child_list->next->node = newnode = new_node();
			newnode->par_count = 1;
			newnode->parent_list = new_node_list(node, NULL);
			newnode->conflicts =
				new_node_set();
			clear_node_set(newnode->conflicts);
			newnode->ancestors =
				new_node_set();
			copy_node_set(node->ancestors, newnode->ancestors);
			SetBit(newnode->ancestors, newnode->id);
			newnode->child_count = countb;
			newnode->child_list = listb;
			update_ancestors(listb, newnode->id);
			replace_parent(listb, node, newnode);
			update_cost(node);
			update_cost(node->child_list->node);
			update_cost(node->child_list->next->node);
		}
	}
}


static void enter_conflict(NODEPTR n1, NODEPTR n2) {

SetBit(n1->conflicts, n2->id);
n1->conflict_list = new_node_list(n2, n1->conflict_list);
n1->conflict_count++;
SetBit(n2->conflicts, n1->id);
n2->conflict_list = new_node_list(n1, n2->conflict_list);
n2->conflict_count++;
}


static void conflict_ancestors(NODEPTR node, NOLIPTR parents) {
NOLIPTR pars;
NODEPTR parnode;

do {
	parnode = parents->node;
	if (TstBit(anc_set, parnode->id)) {
		for (pars = parents->next; pars != NULL; pars = pars->next) {
			parnode = pars->node;
			if (NotTstBit(anc_set, parnode->id)) {
				SetBit(anc_set, parnode->id);
				if (parnode->needs_bit &&
				    NotTstBit(node->conflicts, parnode->id))
					enter_conflict(node, parnode);
				if (parnode->parent_list != NULL)
					conflict_ancestors(node, parnode->parent_list);
				}
			}
		return;
		}
	SetBit(anc_set, parnode->id);
	if (parnode->needs_bit &&
	    NotTstBit(node->conflicts, parnode->id))
		enter_conflict(node, parents->node);
	pars = parents->next; 
	parents = parnode->parent_list;
	while (pars != NULL) {
		parnode = pars->node;
		if (NotTstBit(anc_set, parnode->id)) {
			SetBit(anc_set, parnode->id);
			if (parnode->needs_bit &&
			    NotTstBit(node->conflicts, parnode->id))
				enter_conflict(node, parnode);
			if (parnode->parent_list != NULL)
				conflict_ancestors(node, parnode->parent_list);
			}
		pars = pars->next;
		}
	} while (parents != NULL);
}


static void conflict_childs(NODEPTR node, CNLIPTR childs) {

do {
	if (NotTstBit(node_set, childs->node->id)) {
		SetBit(node_set, childs->node->id);
		if (childs->node->needs_bit &&
		     NotTstBit(node->conflicts, childs->node->id))
		    enter_conflict(node, childs->node);
		if ((childs->node->par_count > 1) &&
		    (NotTstBit(childs->node->ancestors, node->id)))
			conflict_ancestors(node, childs->node->parent_list);
		if (childs->node->child_list != NULL)
			conflict_childs(node, childs->node->child_list);
		}
	childs = childs->next;
	} while (childs != NULL);
}


static void conflict(void) {
NODEPTR node;
NOLIPTR parents;
int     e;

conflict_nodes = e = 0;
uncoloured = NULL;
for (node = root; node != NULL; node = node->next)
	if (node->needs_bit) {             /* expects that root->needs_bit == 0 */
		parents = node->parent_list;
		clear_node_set(node_set);
		clear_node_set(anc_set);
		while (parents != NULL) {
			conflict_childs(node, parents->node->child_list);
			parents = parents->next;
			}
		uncoloured = new_node_list(node, uncoloured);
		conflict_nodes++;
		}
for (node = root; node != NULL; node = node->next)
	e += node->conflict_count;
e = e >> 1;
printf("Conflict graph has %d nodes and %d edges\n", conflict_nodes, e);
avg_degree = (double) e / conflict_nodes;
}


static void sort_min_degree(void) {
int     mindegree;
NOLIPTR nl, prev, min_prev, sort;

for (nl = uncoloured; nl != NULL; nl = nl->next)
	nl->node->help_count = nl->node->conflict_count;
uncoloured = new_node_list(NULL, uncoloured);
sort = NULL;
maxdegree = 0;
while (uncoloured->next != NULL) {
	mindegree = nodecount + 1;
	min_prev = prev = uncoloured;
	for (nl = prev->next; nl != NULL; nl = nl->next) {
		if (nl->node->help_count <= mindegree) {
			min_prev = prev;
			mindegree = nl->node->help_count;
			}
		prev = nl;
		}
	if (maxdegree < mindegree)
		maxdegree = mindegree;
	nl = sort;
	sort = min_prev->next;
	min_prev->next = sort->next;
	sort->next = nl;
	nl = sort->node->conflict_list;
	while (nl != NULL) {
		nl->node->help_count--;
		nl = nl->next;
		}
	}
uncoloured = sort;
}


static void sort_max_degree(void) {
NOLIPTR nl;
SORTPTR sort;
int     i;

for (i = 0; i <= nodecount; i++)
	linear_sort[i].first = NULL;
for (nl = uncoloured; nl != NULL; nl = nl->next) {
	sort = linear_sort + nl->node->conflict_count;
	if (sort->first == NULL)
		sort->first = nl;
	else
		sort->last->next = nl;
	sort->last = nl;
	}
uncoloured = NULL;
for (sort = linear_sort, i = nodecount; i >= 0; sort++, i--)
	if (sort->first != NULL) {
		sort->last->next = uncoloured;
		uncoloured = sort->first;
		}
}


static void sort_path_length(void) {
NOLIPTR nl;
SORTPTR sort;
int     i;

for (i = 0; i <= nodecount; i++)
	linear_sort[i].first = NULL;
for (nl = uncoloured; nl != NULL; nl = nl->next) {
	sort = linear_sort + nl->node->path_length;
	if (sort->first == NULL)
		sort->first = nl;
	else
		sort->last->next = nl;
	sort->last = nl;
	}
uncoloured = NULL;
for (sort = linear_sort, i = nodecount; i >= 0; sort++, i--)
	if (sort->first != NULL) {
		sort->last->next = uncoloured;
		uncoloured = sort->first;
		}
}


static void sort_level(void) {
NOLIPTR nl;
SORTPTR sort;
int     i;

(void) compute_max_level();
for (i = 0; i <= nodecount; i++)
	linear_sort[i].first = NULL;
i -= 2;
for (nl = uncoloured; nl != NULL; nl = nl->next) {
	sort = linear_sort + i - nl->node->level;
	if (sort->first == NULL)
		sort->first = nl;
	else
		sort->last->next = nl;
	sort->last = nl;
	}
uncoloured = NULL;
for (sort = linear_sort, i = nodecount; i >= 0; sort++, i--)
	if (sort->first != NULL) {
		sort->last->next = uncoloured;
		uncoloured = sort->first;
		}
}


static void sort_childs_max_degree(void) {
NODEPTR node, help;
int     maxdegree;
CNLIPTR childs, list, maxlist;

for (node = root; node != NULL; node = node->next)
	if (node->needs_bit && (node->child_count > 1))
		for (childs = node->child_list; childs->next != NULL; childs = childs->next) {
			maxlist = childs;
			maxdegree = childs->node->conflict_count;
			for (list = childs->next; list != NULL; list = list->next)
				if (maxdegree < list->node->conflict_count) {
					maxdegree = list->node->conflict_count;
					maxlist = list;
					}
			help = childs->node;
			childs->node = maxlist->node;
			maxlist->node = help;
			}
}


static void sort_reverse(void) {
NOLIPTR nl, sort;

sort = NULL;
while (uncoloured != NULL) {
	nl = sort;
	sort = uncoloured;
	uncoloured = uncoloured->next;
	sort->next = nl;
	}
uncoloured = sort;
}


static int smallest_free_bit(NODEPTR node) {
NOLIPTR conflicts;
int     i, j;
SETTYPE setword;
SET     conflict_set;

conflicts = node->conflict_list;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = 0; i < SETWORDS; i++) {
	if ((setword = conflict_set[i]) == -1)
		continue;
	j = 0;
	while (setword & 1) {
		setword = setword >> 1;
		j++;
		}
	j = (i << SHIFT) + j;
	if (max_bit_pos < j)
		max_bit_pos = j;
	return j;
	}
system_error("out of bits, increase SETWORDS");
return -1;
}


static int largest_free_bit(NODEPTR node) {
NOLIPTR conflicts;
int     i;
SET     conflict_set;

conflicts = node->conflict_list;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = max_bit_pos; i >= 0; i--)
	if (NotTstBit(conflict_set, i))
		break;
if (i < 0)
	i = ++max_bit_pos;
if (max_bit_pos > MAXSETSIZE)
	system_error("out of bits, increase SETWORDS");
return i;
}


static int max_used_free_bit(NODEPTR node) {
static int use_counts[MAXSETSIZE];
NOLIPTR conflicts;
int     i, maxi, max_use;
SET     conflict_set;

conflicts = node->conflict_list;
if (max_bit_pos == -1)
	for (i = 0; i < MAXSETSIZE; i++)
		use_counts[i] = 0;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = max_bit_pos, max_use = -1; i >= 0; i--)
	if (NotTstBit(conflict_set, i) && (max_use < use_counts[i])) {
		max_use = use_counts[i];
		maxi = i;
		}
if (max_use < 0)
	maxi = ++max_bit_pos;
use_counts[maxi]++;
if (max_bit_pos > MAXSETSIZE)
	system_error("out of bits, increase SETWORDS");
return maxi;
}


static int max_small_last_free_bit(NODEPTR node) {
static int use_counts[MAXSETSIZE];
NOLIPTR conflicts;
int     i, maxi, max_use;
SET     conflict_set;

conflicts = node->conflict_list;
if (max_bit_pos == -1)
	for (i = 0; i < MAXSETSIZE; i++)
		use_counts[i] = 0;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = max_bit_pos, max_use = -1; i >= 0; i--)
	if (NotTstBit(conflict_set, i) && (max_use < use_counts[i])) {
		max_use = use_counts[i];
		maxi = i;
		}
if (max_use < 0)
	maxi = ++max_bit_pos;
use_counts[maxi] += node->help_count;
if (max_bit_pos > MAXSETSIZE)
	system_error("out of bits, increase SETWORDS");
return maxi;
}


static int max_deg_free_bit(NODEPTR node) {
static int use_counts[MAXSETSIZE];
NOLIPTR conflicts;
int     i, maxi, max_use;
SET     conflict_set;

conflicts = node->conflict_list;
if (max_bit_pos == -1)
	for (i = 0; i < MAXSETSIZE; i++)
		use_counts[i] = 0;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = max_bit_pos, max_use = -1; i >= 0; i--)
	if (NotTstBit(conflict_set, i) && (max_use < use_counts[i])) {
		max_use = use_counts[i];
		maxi = i;
		}
if (max_use < 0)
	maxi = ++max_bit_pos;
use_counts[maxi] += node->conflict_count;
if (max_bit_pos > MAXSETSIZE)
	system_error("out of bits, increase SETWORDS");
return maxi;
}


static int last_used_free_bit(NODEPTR node) {
static int use_order[MAXSETSIZE];
NOLIPTR conflicts;
int     i, bitpos;
SET     conflict_set;

conflicts = node->conflict_list;
if (max_bit_pos == -1)
	for (i = 0; i < MAXSETSIZE; i++)
		use_order[i] = 0;
clear_set(conflict_set);
while (conflicts != NULL) {
	if (conflicts->node->bit_pos >= 0) {
		SetBit(conflict_set, conflicts->node->bit_pos);
		}
	conflicts = conflicts->next;
	}
for (i = max_bit_pos; i >= 0; i--)
	if (NotTstBit(conflict_set, use_order[i]))
		break;
if (i < 0)
	bitpos = ++max_bit_pos;
else {
	bitpos = use_order[i];
	while (i < max_bit_pos) {
		use_order[i] = use_order[i+1];
		i++;
		}
	}
use_order[max_bit_pos] = bitpos;
if (max_bit_pos > MAXSETSIZE)
	system_error("out of bits, increase SETWORDS");
return bitpos;
}


static void exchange(NOLIPTR uncol) {
static NODEPTR one_node[MAXSETSIZE+1];
static int node_count = 0;
NODEPTR node, xnode, onenode;
NOLIPTR x, nl;
int     i, max_one_use, node_id, xbit, onebit;
int     *x_col_count;

/* store nodes, the colour of which appears only once, in the one_node array  */

node_count++;
node = uncol->node;
x_col_count = node->col_count;
max_one_use = 0;
for (i = max_bit_pos; i >= 0; i--)
	if (x_col_count[i] < 0)               /* enough free colours, no exchange */
		return;
for (nl = node->conflict_list; nl != NULL; nl = nl->next) 
	if ((xbit = nl->node->bit_pos) >= 0)
		if (x_col_count[xbit] == 0) {
			one_node[max_one_use] = nl->node;
			max_one_use++;
			}
if (max_one_use == 0)                     /* no one used colours, no exchange */
	return;

/* test all nodes which are not in conflict with the new node for exchange    */

/* printf("conflict count of node %4d is %4d, max_one_use is %2d, max_bit_pos is %d\n",
        node_count, node->conflict_count, max_one_use, max_bit_pos); */
node_id = node->id;
for (x = uncoloured; x != uncol; x = x->next) {
	xnode = x->node;
	if (NotTstBit(xnode->conflicts, node_id)) {
		xbit = xnode->bit_pos;
		x_col_count = xnode->col_count;
		for (i = 0; i < max_one_use; i++) {
			onenode = one_node[i];
			onebit = onenode->bit_pos;
			/* the colours of exchange node and one node must be different    */
			/* and are not allowed to be in used by a conlficting node        */

			if ((xbit != onebit) && (x_col_count[onebit] < 0) &&
			    (onenode->col_count[xbit] < 0)) {
				printf("Successful exchange !!!!!!!!!!!!!!!!!!\n");
				xnode->bit_pos = onebit;
				onenode->bit_pos = xbit;

				/* update colour counts */

				for (nl = xnode->conflict_list; nl != NULL; nl = nl->next) {
					nl->node->col_count[xbit]--;
					nl->node->col_count[onebit]++;
					}
				for (nl = onenode->conflict_list; nl != NULL; nl = nl->next) {
					nl->node->col_count[onebit]--;
					nl->node->col_count[xbit]++;
					}
				return;
				}
			}
		}
	}
}


static void colour_exchange(INTNOLIPTRFUNC free_bit) {
NODEPTR node, nlnode;
NOLIPTR uncol, nl;
int     i, bitpos, *col_count;

max_bit_pos = -1;
for (uncol = uncoloured; uncol != NULL; uncol = uncol->next) {
	col_count = uncol->node->col_count;
	uncol->node->bit_pos = -1;
	for (i = 0; i < MAXSETSIZE; i++)
		col_count[i] = -1;                 /* !! -1 means 0; 0 means 1; etc   */
	}


for (uncol = uncoloured; uncol != NULL; uncol = uncol->next) {
	exchange(uncol);
	node = uncol->node;
	node->bit_pos = free_bit(node);
/*	printf("Node %d got color %d\n", node->id, node->bit_pos); */

	/* update colour counts */

	bitpos = node->bit_pos;
	col_count = node->col_count;
	for (nl = node->conflict_list; nl != NULL; nl = nl->next) {
		nlnode = nl->node;
		if (nlnode->bit_pos >= 0)
			col_count[nlnode->bit_pos]++;
		nlnode->col_count[bitpos]++;
		}
	}
printf("Different bits needed: %d\n", max_bit_pos + 1);
}


static void colour(INTNOLIPTRFUNC free_bit) {
NOLIPTR uncol;

max_bit_pos = -1;
for (uncol = uncoloured; uncol != NULL; uncol = uncol->next)
	uncol->node->bit_pos = -1;
for (uncol = uncoloured; uncol != NULL; uncol = uncol->next)
	uncol->node->bit_pos = free_bit(uncol->node);
printf("Colour bits needed: %d\n", max_bit_pos + 1);
}


void colour_parents(INTNOLIPTRFUNC free_bit, NOLIPTR parents) {
while (parents != NULL) {
	if (parents->node->bit_pos == -1) {
		colour_parents(free_bit, parents->node->parent_list);
		parents->node->bit_pos = -2;
		if (parents->node->needs_bit)
			parents->node->bit_pos = free_bit(parents->node);
		}
	parents = parents->next;
	}
}


static void colour_childs(INTNOLIPTRFUNC free_bit, CNLIPTR childs) {
while (childs != NULL) {
	if (NotTstBit(node_set, childs->node->id)) {
		SetBit(node_set, childs->node->id);
		if (childs->node->bit_pos == -1) {
			colour_parents(free_bit, childs->node->parent_list);
			childs->node->bit_pos = -2;
			if (childs->node->needs_bit)
				childs->node->bit_pos = free_bit(childs->node);
			}
		colour_childs(free_bit, childs->node->child_list);
		}
	childs = childs->next;
	}
}


static void colour_caseau(INTNOLIPTRFUNC free_bit) {
NODEPTR node;

max_bit_pos = -1;
for (node = root; node != NULL; node = node->next)
	node->bit_pos = -1;
clear_node_set(node_set);
colour_childs(free_bit, root->child_list);
printf("Different bits needed: %d\n", max_bit_pos + 1);
}


static void colour_down(INTNOLIPTRFUNC free_bit, CNLIPTR childs) {
while (childs != NULL) {
	if (childs->node->bit_pos == -1) {
		childs->node->bit_pos = -2;
		if (childs->node->needs_bit)
			childs->node->bit_pos = free_bit(childs->node);
		colour_down(free_bit, childs->node->child_list);
		}
	childs = childs->next;
	}
}


static void colour_top_down(INTNOLIPTRFUNC free_bit) {
NODEPTR node;

max_bit_pos = -1;
for (node = root; node != NULL; node = node->next)
	node->bit_pos = -1;
colour_down(free_bit, root->child_list);
printf("Different bits needed: %d\n", max_bit_pos + 1);
}


static void code_parent(SET s, NOLIPTR parents) {
int     bp;

while (parents != NULL) {
	if (NotTstBit(anc_set, parents->node->id)) {
		SetBit(anc_set, parents->node->id);
		if ((bp = parents->node->bit_pos) >= 0) {
			SetBit(s, bp);
			}
		code_parent(s, parents->node->parent_list);
		}
	parents = parents->next;
	}
}


static void code(NODEPTR node) {
int     bp;

clear_node_set(anc_set);
clear_set(node->bit_set);
if ((bp = node->bit_pos) >= 0) {
	SetBit(node->bit_set, bp);
	}
code_parent(node->bit_set, node->parent_list);
}


static print_code(SET bit_set) {
int     i, j, k;
SETTYPE setword;

k = 0;
for (i = 0; i < SETWORDS; i++) {
	if ((setword = bit_set[i]) == 0) {
		k += MASK + 1;
		continue;
		}
	for (j = 0; j <= MASK; j++) {
		if (setword & 1)
			printf(" %d", k);
		setword = setword >> 1;
		k++;
		}
	}
}


static void print_hierarchy(void) {
int     i;
NODEPTR node;
NOLIPTR pars;

for (i = 1; i <= max_level; i++) {
	for (node = root; node != NULL; node = node->next)
		if (node->level == i) {
			if (strcmp(node->name, "") == 0)
				fprintf(stderr, "%d(", node->id);
			else
				fprintf(stderr, "%s(", node->name);
			for (pars = node->parent_list; pars != NULL; pars = pars->next) {
				if (strcmp(pars->node->name, "") == 0)
					fprintf(stderr, "%d", pars->node->id);
				else
					fprintf(stderr, "%s", pars->node->name);
				if (pars->next != NULL)
					fprintf(stderr, ",");
				}
			fprintf(stderr, ") ");
			}
	fprintf(stderr, "\n");
	}
}


static void test_codes(void) {
NODEPTR i, j;
int     n, m, ishift, is_anc;
SETPTR  jbits, ibits;
SETTYPE imask, ibit0;

for (i = root, n = orgnodecount; n > 0; n--, i = i->next)
	code(i);
for (i = root, n = orgnodecount; n > 0; n--, i = i->next) {
	ishift = i->id >> SHIFT;
	imask = ((SETTYPE) 1) << (i->id & MASK);
	ibits = i->bit_set;
	ibit0 = ibits[0];
	for (j = root, m = orgnodecount; m > 0; m--, j = j->next) {
		jbits = j->bit_set;
		is_anc = TstBitShiftMask(j->ancestors, ishift, imask);
		if ((jbits[0] & ibit0) != ibit0) {
			if (is_anc == 0)
				continue;
			}
		else if (a_and_b_is_equal_b_set_1(jbits, ibits) == is_anc)
		    continue;
		printf("Node %3d %17s with set {", j->id, j->name);
		print_code(j->bit_set);
		printf("} and parent node %3d %17s with {", i->id, i->name);
		print_code(i->bit_set);
		printf("} should ");
		if (TstBit(j->ancestors, i->id))
			printf("be in the relation\n");
		else
			printf("not be in the relation\n");
		}
	}
}


int main(int argc, char *argv[]) {
int spliteight, splitnew, splitfast, simple_col, test, x_change, max_first;
int caseau, top_down, briggs, maxbriggs, largest, sort_path, sortlevel, reverse;
int out_hier, splitsix, splitnine;
NODEPTR node;

if ((argc < 2) || (argc > 3)) {
	printf("usage: %s [abcdfhlnmprstwx8] file !!!\n", argv[0]);
	printf("a: give All mi nodes a bit\n");
	printf("b: Briggs colouring\n");
	printf("c: Caseau colouring\n");
	printf("d: Degree colouring\n");
	printf("e: use Early presplitting\n");
	printf("f: use 'Fast' splitting\n");
	printf("g: use Early presplitting\n");
	printf("h: sort secondary Hierachical level before colouring\n");
	printf("l: use additionally Largest free bit\n");
	printf("m: maximum first secondary ordering\n");
	printf("n: use 'New' splitting\n");
	printf("o: Output type hierarchy to stderr\n");
	printf("p: sort secondary for Path length before colouring\n");
	printf("r: reverse before colouring\n");
	printf("s: use Simple graph colouring\n");
	printf("t: Test the generated codes\n");
	printf("u: do not use bUcket presplitting\n");
	printf("v: do not use bVcket presplitting\n");
	printf("w: top doWn colouring\n");
	printf("x: interchange colour technique\n");
	printf("z: new split with old propagation\n");
	printf("6: split lists which are greater or equal than 6\n");
	printf("8: split lists which are greater or equal than 8\n");
	printf("9: split lists which are greater or equal than 9\n");
	exit(0);
	}

splitnine = 0;
spliteight = 0;
splitsix = 0;
splitnew = 0;
splitfast = 0;
z_split = 0;
simple_col = 0;
briggs = 0;
maxbriggs = 0;
test = 0;
all_mi_get_bit = 0;
caseau = 0;
top_down = 0;
largest = 0;
sort_path = 0;
sortlevel = 0;
x_change = 0;
max_first = 0;
reverse = 0;
out_hier = 0;
not_early = 1;
not_bucket = 1;
not_bucket_v = 1;
if (argc == 2)
	init(argv[1]);
else {
	if (strchr(argv[1], 'a') != NULL)
		all_mi_get_bit = 1;
	if (strchr(argv[1], 'b') != NULL)
		briggs = 1;
	if (strchr(argv[1], 'c') != NULL)
		caseau = 1;
	if (strchr(argv[1], 'd') != NULL)
		maxbriggs = 1;
	if (strchr(argv[1], 'e') != NULL)
		not_early = 0;
	if (strchr(argv[1], 'f') != NULL)
		splitfast = 1;
	if (strchr(argv[1], 'h') != NULL)
		sortlevel = 1;
	if (strchr(argv[1], 'l') != NULL)
		largest = 1;
	if (strchr(argv[1], 'n') != NULL)
		splitnew = 1;
	if (strchr(argv[1], 'm') != NULL)
		max_first = 1;
	if (strchr(argv[1], 'o') != NULL)
		out_hier = 1;
	if (strchr(argv[1], 'p') != NULL)
		sort_path = 1;
	if (strchr(argv[1], 'r') != NULL)
		reverse = 1;
	if (strchr(argv[1], 's') != NULL)
		simple_col = 1;
	if (strchr(argv[1], 't') != NULL)
		test = 1;
	if (strchr(argv[1], 'u') != NULL)
		not_bucket = 0;
	if (strchr(argv[1], 'v') != NULL)
		not_bucket_v = 0;
	if (strchr(argv[1], 'w') != NULL)
		top_down = 1;
	if (strchr(argv[1], 'x') != NULL)
		x_change = 1;
	if (strchr(argv[1], 'z') != NULL)
		z_split = 1;
	if (strchr(argv[1], '6') != NULL)
		splitsix = 1;
	if (strchr(argv[1], '8') != NULL)
		spliteight = 1;
	if (strchr(argv[1], '9') != NULL)
		splitnine = 1;
	init(argv[2]);
	}
if (!briggs && !maxbriggs && !caseau && !top_down)
	simple_col = 1;
read_data();
build_matrix();
if (out_hier)
	print_hierarchy();
matrix_bytes = (((nodecount + 31) >> 5) << 2) * nodecount;
printf("The number of bytes in the binary matrix is          %6d(1.0000)\n",
       matrix_bytes);
{
int byte_count, i;
for (i = 1, byte_count = 0; i <= nodecount; i++) {
	byte_count += (((i + 31) >> 5) << 2);
	}
printf("The number of bytes in the trimmed binary matrix is  %6d(%6.4f)\n",
       byte_count, (double) byte_count / matrix_bytes);
}
printf("The number of bytes in the ancestor array is         %6d(%6.4f)\n",
       (bit_count + nodecount) << 1, (double) ((bit_count + nodecount) << 1) / matrix_bytes);
printf("The number of bytes in the ancestor list is          %6d(%6.4f)\n",
       anc_list_bytes, (double) anc_list_bytes / matrix_bytes);
printf("The number of bytes in the minimal representation is %6d(%6.4f)\n", min_bytes,
        (double) min_bytes / matrix_bytes);
if (splitnew) {
	compute_bounds();
	pre_split();
	new_splits();
	}
else if (splitfast) {
	compute_bounds();
	pre_split();
	fast_splits();
	}
else if (splitsix) {
	compute_bounds();
	eight_split(6);
	compute_path_length();
	}
else if (spliteight) {
	compute_bounds();
	eight_split(8);
	compute_path_length();
	}
else if (splitnine) {
	compute_bounds();
	eight_split(9);
	compute_path_length();
	}
conflict();
if (sortlevel)
	sort_level();
if (sort_path) {
	sort_path_length();
	if (caseau || top_down)
		sort_childs_max_degree();
	}
if (max_first)
	sort_max_degree();
if (reverse)
	sort_reverse();
if (simple_col) {
	printf("\n");
	if (largest) {
		printf("Simple colouring (largest_free_bit):  ");
		colour(largest_free_bit);
		}
	printf("Simple colouring (last_use_free_bit): ");
	colour(last_used_free_bit);
	printf("Simple colouring (max_used_free_bit): ");
	colour(max_used_free_bit);
	printf("Simple colouring (smallest_free_bit): ");
	colour(smallest_free_bit);
	printf("Simple colouring (max_deg_free_bit):  ");
	colour(max_deg_free_bit);
	if (x_change) {
		printf("Simple exchange colouring (smallest_free_bit): ");
		colour_exchange(smallest_free_bit);
		printf("Simple exchange colouring (max_deg_free_bit):  ");
		colour_exchange(max_deg_free_bit);
		}
	}
if (briggs) {
	printf("\n");
	sort_min_degree();
	if (largest) {
		printf("Briggs colouring (largest_free_bit):  ");
		colour(largest_free_bit);
		}
	printf("Briggs colouring (last_use_free_bit): ");
	colour(last_used_free_bit);
	printf("Briggs colouring (max_sdl_free_bit):  ");
	colour(max_small_last_free_bit);
	printf("Briggs colouring (max_used_free_bit): ");
	colour(max_used_free_bit);
	printf("Briggs colouring (smallest_free_bit): ");
	colour(smallest_free_bit);
	printf("Briggs colouring (max_deg_free_bit):  ");
	colour(max_deg_free_bit);
	if (x_change) {
		printf("Briggs exchange colouring (smallest_free_bit): ");
		colour_exchange(smallest_free_bit);
		printf("Briggs exchange colouring (max_deg_free_bit):  ");
		colour_exchange(max_deg_free_bit);
		}
	}
if (maxbriggs) {
	printf("\n");
	sort_max_degree();
	if (largest) {
		printf("Degree colouring (largest_free_bit):  ");
		colour(largest_free_bit);
		}
	printf("Degree colouring (last_use_free_bit): ");
	colour(last_used_free_bit);
	printf("Degree colouring (max_used_free_bit): ");
	colour(max_used_free_bit);
	printf("Degree colouring (smallest_free_bit): ");
	colour(smallest_free_bit);
	printf("Degree colouring (max_deg_free_bit):  ");
	colour(max_deg_free_bit);
	if (x_change) {
		printf("Degree exchange colouring (smallest_free_bit): ");
		colour_exchange(smallest_free_bit);
		printf("Degree exchange colouring (max_deg_free_bit):  ");
		colour_exchange(max_deg_free_bit);
		}
	}
if (caseau) {
	printf("\n");
	if (largest) {
		printf("Caseau colouring (largest_free_bit):  ");
		colour_caseau(largest_free_bit);
		}
	printf("Caseau colouring (last_use_free_bit): ");
	colour_caseau(last_used_free_bit);
	printf("Caseau colouring (max_used_free_bit): ");
	colour_caseau(max_used_free_bit);
	printf("Caseau colouring (smallest_free_bit): ");
	colour_caseau(smallest_free_bit);
	printf("Caseau colouring (max_deg_free_bit):  ");
	colour_caseau(max_deg_free_bit);
	}
if (top_down) {
	printf("\n");
	if (largest) {
		printf("  Down colouring (largest_free_bit):  ");
		colour_top_down(largest_free_bit);
		}
	printf("  Down colouring (last_use_free_bit): ");
	colour_top_down(last_used_free_bit);
	printf("  Down colouring (max_used_free_bit): ");
	colour_top_down(max_used_free_bit);
	printf("  Down colouring (smallest_free_bit): ");
	colour_top_down(smallest_free_bit);
	printf("  Down colouring (max_deg_free_bit):  ");
	colour_top_down(max_deg_free_bit);
	}
if (test)
	test_codes();
printf("The number of bytes for colouring is                 %6d(%6.4f)\n",
       (((max_bit_pos + 32) >> 5) << 2) * orgnodecount,
       (double) ((((max_bit_pos + 31) >> 5) << 2) * orgnodecount) / matrix_bytes);
printf("Different bits needed: %d\n", max_bit_pos + 1);
printf("The biggest degree is %d\n", maxdegree);
printf("The lower bound for the path length was    %3d\n", root->lower_bound);
compute_bounds();
printf("The lower bound for the path length is now %3d\n", root->lower_bound);
printf("The path length is                         %3d\n", root->path_length);
printf("The average degree is                    %5.2f\n", avg_degree);
printf("%10.3f kbytes memory wasted during computation\n", (double) mem_use /1024);
}
